-------------------------------------------------------------
--	Filename:  HALFBANDDECIMATIONFILT.VHD
-- Authors:   Alain Zarembowitch 	 
--	Version: 1
--	Date last modified: 2/28/09
-- Inheritance: 	HALFBANDFILT.VHD 2/24/09, ANTIALIASING2.VHD 8-10-01
--
-- Description: Half-band filter. Simple 10th order FIR filter 
-- designed to have flat (+/0.1 dB) gain in the +/- 0.15*fs passband, and linear phase of course (FIR),
-- where fs is the sampling frequency. 
-- The filter aimed at rejecting aliasing at the modulator output.
-- The filter structure is simple so that several of these filters can be concatenated for small data rates. 
-- Rejection between 0.3 and 0.5*fs is 40dB or better
-- The filter decimates the output by 1:2
-- Two filter is shared by two channels.
--
-- Implementation: 
-- Synthesized using Matlab's Filter design toolbox.
-- Coefficients are then modified to minimize filter complexity.
-- Filter coefficients (after quantization and minimizing the number of 1's)
-- b = [20,0,-72,0,312,512,312,0,-72,0,20]
-- For the actual frequency response, use freqz(b)
---------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
--library UNISIM;
--use UNISIM.VComponents.all;

entity HALFBANDDECIMATIONFILT is
    port ( 

	--// GLOBAL CLOCKS, RESET
   CLK : in std_logic;				-- Master clock for this FPGA
	SYNC_RESET: in std_logic;		-- reset
	
	SAMPLE_CLK_IN: in std_logic;
		-- read input sample at rising_edge of CLK and SAMPLE_CLK_IN  = '1'.
		-- One CLK wide pulse. Maximum rate: 1 input sample per CLK.
	DATA_I_IN: in std_logic_vector(13 downto 0);
	DATA_Q_IN: in std_logic_vector(13 downto 0);
		-- input signal. 2's complement.
		-- read at rising edge of CLK when SAMPLE_CLK_IN = '1'

	SAMPLE_CLK_OUT: out std_logic;
		-- half-sampling rate. 
		-- latency: 1 sample + 5 CLKs
	DATA_I_OUT: out std_logic_vector(13 downto 0);
	DATA_Q_OUT: out std_logic_vector(13 downto 0);
		-- output signals. 2's complement.
		-- Read at rising edge of CLK and SAMPLE_CLK_OUT = '1'.
	SATURATION: out std_logic
		-- '1' when a saturation occurs during filter computation
			);
end entity;

architecture behavioral of HALFBANDDECIMATIONFILT is
--------------------------------------------------------
--      COMPONENTS
--------------------------------------------------------

--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- Suffix _D indicates a one CLK delayed version of the net with the same name
-- Suffix _E indicates an extended precision version of the net with the same name
-- Suffix _N indicates an inverted version of the net with the same name
-- Suffix _INC indicates increment.
signal SAMPLE_CLK_IN_D: std_logic;
signal SAMPLE_CLK_IN_D2: std_logic;
signal SAMPLE_CLK_IN_D3: std_logic;
signal SAMPLE_CLK_IN_D4: std_logic;
signal SAMPLE_COUNT: std_logic_vector(0 downto 0);
signal SAMPLE_COUNT_D: std_logic_vector(0 downto 0);
signal SAMPLE_COUNT_D2: std_logic_vector(0 downto 0);
signal SAMPLE_COUNT_D3: std_logic_vector(0 downto 0);

signal FIL_I_IN0: std_logic_vector(13 downto 0);
signal FIL_I_IN1: std_logic_vector(13 downto 0);
signal FIL_I_IN2: std_logic_vector(13 downto 0);
signal FIL_I_IN3: std_logic_vector(13 downto 0);
signal FIL_I_IN4: std_logic_vector(13 downto 0);
signal FIL_I_IN5: std_logic_vector(13 downto 0);
signal FIL_I_IN6: std_logic_vector(13 downto 0);
signal FIL_I_IN7: std_logic_vector(13 downto 0);
signal FIL_I_IN8: std_logic_vector(13 downto 0);
signal FIL_I_IN9: std_logic_vector(13 downto 0);
signal FIL_I_IN10: std_logic_vector(13 downto 0);
signal FIL_Q_IN0: std_logic_vector(13 downto 0);
signal FIL_Q_IN1: std_logic_vector(13 downto 0);
signal FIL_Q_IN2: std_logic_vector(13 downto 0);
signal FIL_Q_IN3: std_logic_vector(13 downto 0);
signal FIL_Q_IN4: std_logic_vector(13 downto 0);
signal FIL_Q_IN5: std_logic_vector(13 downto 0);
signal FIL_Q_IN6: std_logic_vector(13 downto 0);
signal FIL_Q_IN7: std_logic_vector(13 downto 0);
signal FIL_Q_IN8: std_logic_vector(13 downto 0);
signal FIL_Q_IN9: std_logic_vector(13 downto 0);
signal FIL_Q_IN10: std_logic_vector(13 downto 0);
signal FIL_Q_IN11: std_logic_vector(13 downto 0);


signal FIL_IN0: std_logic_vector(13 downto 0);
signal FIL_IN1: std_logic_vector(13 downto 0);
signal FIL_IN2: std_logic_vector(13 downto 0);
signal FIL_IN3: std_logic_vector(13 downto 0);
signal FIL_IN4: std_logic_vector(13 downto 0);
signal FIL_IN5: std_logic_vector(13 downto 0);
signal FIL_IN6: std_logic_vector(13 downto 0);
signal FIL_IN7: std_logic_vector(13 downto 0);
signal FIL_IN8: std_logic_vector(13 downto 0);
signal FIL_IN9: std_logic_vector(13 downto 0);
signal FIL_IN10: std_logic_vector(13 downto 0);

signal FIL_A: std_logic_vector(11 downto 0);
signal FIL_A_D: std_logic_vector(11 downto 0);
signal FIL_A1: std_logic_vector(11 downto 0);
signal FIL_A2: std_logic_vector(11 downto 0);
signal FIL_B: std_logic_vector(13 downto 0);
signal FIL_B_D: std_logic_vector(13 downto 0);
signal FIL_B1: std_logic_vector(13 downto 0);
signal FIL_B2: std_logic_vector(13 downto 0);
signal FIL_C: std_logic_vector(15 downto 0);
signal FIL_C_D: std_logic_vector(15 downto 0);
signal FIL_C1: std_logic_vector(15 downto 0);
signal FIL_C2: std_logic_vector(15 downto 0);
signal FIL_D: std_logic_vector(16 downto 0);
signal FIL_D_D: std_logic_vector(16 downto 0);
signal FIL_D1: std_logic_vector(16 downto 0);
signal FIL_D2: std_logic_vector(16 downto 0);
signal FIL_E: std_logic_vector(11 downto 0);
signal FIL_E_D: std_logic_vector(11 downto 0);
signal FIL_E1: std_logic_vector(11 downto 0);
signal FIL_E2: std_logic_vector(11 downto 0);
signal FIL_F: std_logic_vector(12 downto 0);
signal FIL_F1: std_logic_vector(12 downto 0);
signal FIL_F2: std_logic_vector(12 downto 0);
signal FIL_G: std_logic_vector(17 downto 0);
signal FIL_G_D: std_logic_vector(17 downto 0);
signal FIL_G1: std_logic_vector(17 downto 0);
signal FIL_G2: std_logic_vector(17 downto 0);
signal FIL_H: std_logic_vector(14 downto 0);
signal FIL_H1: std_logic_vector(14 downto 0);
signal FIL_H2: std_logic_vector(14 downto 0);
signal FIL_I: std_logic_vector(15 downto 0);
signal FIL_I_D: std_logic_vector(15 downto 0);
signal FIL_I1: std_logic_vector(15 downto 0);
signal FIL_I2: std_logic_vector(15 downto 0);
signal FIL_J: std_logic_vector(18 downto 0);
signal FIL_J1: std_logic_vector(18 downto 0);
signal FIL_J2: std_logic_vector(18 downto 0);
signal FIL_K: std_logic_vector(13 downto 0);
signal DATA_I_OUT_TEMP: std_logic_vector(13 downto 0);

--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin


--// STORE INPUT SAMPLES ---------------------
SHIFT_001: process(CLK)
begin
	if(rising_edge(CLK)) then
		SAMPLE_CLK_IN_D <= SAMPLE_CLK_IN;
		
		if(SAMPLE_CLK_IN = '1') then	
			SAMPLE_COUNT <= SAMPLE_COUNT + 1;  -- modulo 2

			FIL_I_IN0 <= DATA_I_IN;                                                
			FIL_I_IN1 <= FIL_I_IN0;                                                
			FIL_I_IN2 <= FIL_I_IN1;                                                
			FIL_I_IN3 <= FIL_I_IN2;                                                
			FIL_I_IN4 <= FIL_I_IN3;                                                
			FIL_I_IN5 <= FIL_I_IN4;                                                
			FIL_I_IN6 <= FIL_I_IN5;                                                
			FIL_I_IN7 <= FIL_I_IN6;                                                
			FIL_I_IN8 <= FIL_I_IN7;                                                
			FIL_I_IN9 <= FIL_I_IN8;                                                
			FIL_I_IN10 <= FIL_I_IN9;      

			-- register is one sample longer than for I as Q is processed one sample later
			FIL_Q_IN0 <= DATA_Q_IN;                                                
			FIL_Q_IN1 <= FIL_Q_IN0;                                                
			FIL_Q_IN2 <= FIL_Q_IN1;                                                
			FIL_Q_IN3 <= FIL_Q_IN2;                                                
			FIL_Q_IN4 <= FIL_Q_IN3;                                                
			FIL_Q_IN5 <= FIL_Q_IN4;                                                
			FIL_Q_IN6 <= FIL_Q_IN5;                                                
			FIL_Q_IN7 <= FIL_Q_IN6;                                                
			FIL_Q_IN8 <= FIL_Q_IN7;                                                
			FIL_Q_IN9 <= FIL_Q_IN8;                                                
			FIL_Q_IN10 <= FIL_Q_IN9;                                                
			FIL_Q_IN11 <= FIL_Q_IN10;                                                
		end if;
	end if;
end process;

-- The filter engine is shared between two I and Q streams
-- Multiplex the I/Q data streams here
-- Take earlier Q sample, as Q channel processing is done one sample after I.
MUX001: process(SAMPLE_COUNT, FIL_I_IN0, FIL_I_IN1, FIL_I_IN2, FIL_I_IN3, FIL_I_IN4, FIL_I_IN5,
	FIL_I_IN6, FIL_I_IN7, FIL_I_IN8, FIL_I_IN9, FIL_I_IN10, FIL_Q_IN1, FIL_Q_IN2,
	FIL_Q_IN3, FIL_Q_IN4, FIL_Q_IN5, FIL_Q_IN6, FIL_Q_IN7, FIL_Q_IN8, FIL_Q_IN9, FIL_Q_IN10, FIL_Q_IN11)
begin       
	if(SAMPLE_COUNT = 0) then
		FIL_IN0 <= FIL_I_IN0;
		FIL_IN1 <= FIL_I_IN1;
		FIL_IN2 <= FIL_I_IN2;
		FIL_IN3 <= FIL_I_IN3;
		FIL_IN4 <= FIL_I_IN4;
		FIL_IN5 <= FIL_I_IN5;
		FIL_IN6 <= FIL_I_IN6;
		FIL_IN7 <= FIL_I_IN7;
		FIL_IN8 <= FIL_I_IN8;
		FIL_IN9 <= FIL_I_IN9;
		FIL_IN10 <= FIL_I_IN10;
	else
		FIL_IN0 <= FIL_Q_IN1;
		FIL_IN1 <= FIL_Q_IN2;
		FIL_IN2 <= FIL_Q_IN3;
		FIL_IN3 <= FIL_Q_IN4;
		FIL_IN4 <= FIL_Q_IN5;
		FIL_IN5 <= FIL_Q_IN6;
		FIL_IN6 <= FIL_Q_IN7;
		FIL_IN7 <= FIL_Q_IN8;
		FIL_IN8 <= FIL_Q_IN9;
		FIL_IN9 <= FIL_Q_IN10;
		FIL_IN10 <= FIL_Q_IN11;
	end if;	
end process;    


-- FIRST LEVEL OF ADDERS (fold)
-- coeff(1) = coeff(11) = 16/1024 -> shift right 6-3 (for better computation precision)
FIL_A1 <= FIL_IN0(13) & FIL_IN0(13 downto 3);
FIL_A2 <= FIL_IN10(13) & FIL_IN10(13 downto 3);
FIL_A <= FIL_A1 + FIL_A2;

-- coeff(3) = coeff(9) = 64/1024 -> shift right 4-3
FIL_B1 <= FIL_IN2(13) & FIL_IN2(13 downto 1);
FIL_B2 <= FIL_IN8(13) & FIL_IN8(13 downto 1);
FIL_B <= FIL_B1 + FIL_B2;

-- coeff(5) = coeff(7) = 256/1024 -> shift right 2 - 3
FIL_C1 <= FIL_IN4(13) & FIL_IN4(13 downto 0) & '0';
FIL_C2 <= FIL_IN6(13) & FIL_IN6(13 downto 0) & '0';
FIL_C <= FIL_C1 + FIL_C2;

-- SECOND LEVEL OF ADDERS
-- coeff(6) (512/1024) + coeff(5/7) (64/1024)
FIL_D1 <= FIL_IN5(13) & FIL_IN5(13 downto 0) & "00";
FIL_D2(13 downto 0) <= FIL_C(15 downto 2);
FIL_D2(16 downto 14) <= (others => FIL_C(15));
FIL_D <= FIL_D1 + FIL_D2;

-- coeff(3/9) (8/1024) + coeff(5/7) (8/1024)
FIL_E1 <= FIL_B(13) & FIL_B(13 downto 3);
FIL_E2 <= FIL_C(15) & FIL_C(15 downto 5);
FIL_E <= FIL_E1 + FIL_E2;

-- THIRD LEVEL OF ADDERS  
RECLOCK_001: process(CLK)
begin   
	if(rising_edge(CLK)) then
		SAMPLE_CLK_IN_D2 <= SAMPLE_CLK_IN_D;
		SAMPLE_COUNT_D <= SAMPLE_COUNT;
		
		if(SAMPLE_CLK_IN_D = '1') then	
			FIL_A_D <= FIL_A;
			FIL_B_D <= FIL_B;
			FIL_C_D <= FIL_C;
			FIL_D_D <= FIL_D;
			FIL_E_D <= FIL_E;
		end if;
	end if;
end process;

-- FIL_A_D - FIL_E_D
FIL_F1 <= FIL_A_D(11) & FIL_A_D(11 downto 0);
FIL_F2 <= FIL_E_D(11) & FIL_E_D(11 downto 0);
FIL_F <= FIL_F1 - FIL_F2;

-- FIL_C_D + FIL_D_D
FIL_G1 <= FIL_C_D(15) & FIL_C_D(15) & FIL_C_D(15 downto 0);
FIL_G2 <= FIL_D_D(16) & FIL_D_D(16 downto 0);
FIL_G <= FIL_G1 + FIL_G2;

-- FIL_A_D/4 - FIL_B_D
FIL_H1(9 downto 0) <= FIL_A_D(11 downto 2);
FIL_H1(14 downto 10) <= (others => FIL_A_D(11));
FIL_H2 <= FIL_B_D(13) & FIL_B_D(13 downto 0);
FIL_H <= FIL_H1 - FIL_H2;

-- FORTH LEVEL OF ADDERS  
-- FIL_F + FIL_H
FIL_I1(12 downto 0) <= FIL_F(12 downto 0);
FIL_I1(15 downto 13) <= (others => FIL_F(12));
FIL_I2 <= FIL_H(14) & FIL_H(14 downto 0);
FIL_I <= FIL_I1 + FIL_I2;


RECLOCK_002: process(CLK)
begin   
	if(rising_edge(CLK)) then
		SAMPLE_CLK_IN_D3 <= SAMPLE_CLK_IN_D2;
		SAMPLE_COUNT_D2 <= SAMPLE_COUNT_D;
		
		if(SAMPLE_CLK_IN_D2 = '1') then	
			FIL_G_D <= FIL_G;
			FIL_I_D <= FIL_I;
		end if;
	end if;
end process;

-- FIFTH LEVEL OF ADDERS
-- FIL_G_D + FIL_I_D  
FIL_J1 <= FIL_G_D(17) & FIL_G_D(17 downto 0);
FIL_J2(15 downto 0) <= FIL_I_D(15 downto 0);
FIL_J2(18 downto 16) <= (others => FIL_I_D(15));
FIL_J <= FIL_J1 + FIL_J2;


-- clip output to 13 bits (same number of bits as input)  
-- Should be OK since the filter has a maximum gain of 1.
CLIP001: process(CLK)
begin                         
	if rising_edge(CLK) then
		SAMPLE_CLK_IN_D4 <= SAMPLE_CLK_IN_D3;
		SAMPLE_COUNT_D3 <= SAMPLE_COUNT_D2;

		if(SAMPLE_CLK_IN_D3 = '1') then	
			if(FIL_J(18) = '1') and (FIL_J(17 downto 16) /= "11") then	-- overflow negative
				FIL_K <= "10000000000001";	-- most negative
				SATURATION <= '1';
			elsif(FIL_J(18) = '0') and (FIL_J(17 downto 16) /= "00") then	-- overflow positive
				FIL_K <= "01111111111111";	-- most positive
				SATURATION <= '1';
			else	-- within range
				FIL_K <= FIL_J(16 downto 3);
				SATURATION <= '0';
			end if;		
		end if;
	end if;
end process;


GET_OUTPUT_SAMPLES_001: process(CLK)
begin                         
	if rising_edge(CLK) then
		if (SAMPLE_CLK_IN_D4 = '1') then
			if (SAMPLE_COUNT_D3 = 0) then
				-- save I output until we compute the matching Q output
				DATA_I_OUT_TEMP <= FIL_K;
				SAMPLE_CLK_OUT <= '0';
			else
				-- output I and Q together
				DATA_Q_OUT <= FIL_K;
				DATA_I_OUT <= DATA_I_OUT_TEMP;
				SAMPLE_CLK_OUT <= '1';
			end if;
	  	else
			SAMPLE_CLK_OUT <= '0';
		end if;
	end if;
end process;

		

end Behavioral;
